@@ -24,7 +24,7 @@ import os
 from os.path import dirname, exists, expanduser
 import re
 import shlex
-from subprocess import Popen, PIPE
+from subprocess import call, Popen, PIPE, DEVNULL
 import sys
 from threading import Thread
 from time import sleep
@@ -91,38 +91,89 @@ def splittext(text, search, attr):
 
 class URLChooser:
 
-    def __init__(self, extractedurls, compact=False, dedupe=False, shorten=True,
-                 run="", pipe=False):
+    def __init__(self, extractedurls, compact=False, nohelp=False, dedupe=False,
+                 shorten=True, run="", pipe=False, genconf=False):
         self.conf = expanduser("~/.config/urlscan/config.json")
-        self.palettes = []
+        self.keys = {'/': self._search_key,
+                     '0': self._digits,
+                     '1': self._digits,
+                     '2': self._digits,
+                     '3': self._digits,
+                     '4': self._digits,
+                     '5': self._digits,
+                     '6': self._digits,
+                     '7': self._digits,
+                     '8': self._digits,
+                     '9': self._digits,
+                     'C': self._clipboard,
+                     'c': self._context,
+                     'ctrl l': self._clear_screen,
+                     'f1': self._help_menu,
+                     'G': self._bottom,
+                     'g': self._top,
+                     'j': self._down,
+                     'k': self._up,
+                     'P': self._clipboard_pri,
+                     'l': self._link_handler,
+                     'p': self._palette,
+                     'Q': self._quit,
+                     'q': self._quit,
+                     'S': self._all_shorten,
+                     's': self._shorten,
+                     'u': self._all_escape
+                     }
+        self.palettes = {}
+        # Default color palette
+        default = [('header', 'white', 'dark blue', 'standout'),
+                   ('footer', 'white', 'dark red', 'standout'),
+                   ('search', 'white', 'dark green', 'standout'),
+                   ('msgtext', '', ''),
+                   ('msgtext:ellipses', 'light gray', 'black'),
+                   ('urlref:number:braces', 'light gray', 'black'),
+                   ('urlref:number', 'yellow', 'black', 'standout'),
+                   ('urlref:url', 'white', 'black', 'standout'),
+                   ('url:sel', 'white', 'dark blue', 'bold')]
+        # Default black & white palette
+        blw = [('header', 'black', 'light gray', 'standout'),
+               ('footer', 'black', 'light gray', 'standout'),
+               ('search', 'black', 'light gray', 'standout'),
+               ('msgtext', '', ''),
+               ('msgtext:ellipses', 'white', 'black'),
+               ('urlref:number:braces', 'white', 'black'),
+               ('urlref:number', 'white', 'black', 'standout'),
+               ('urlref:url', 'white', 'black', 'standout'),
+               ('url:sel', 'black', 'light gray', 'bold')]
+        self.palettes.update([("default", default), ("bw", blw)])
+        if genconf is True:
+            self._config_create()
         try:
             with open(self.conf, 'r') as conf_file:
                 data = json.load(conf_file)
-                for pal in data.values():
-                    self.palettes.append([tuple(i) for i in pal])
+                try:
+                    for pal_name, pal in data['palettes'].items():
+                        self.palettes.update([(pal_name, [tuple(i) for i in pal])])
+                except KeyError:
+                    pass
+                try:
+                    items = data['keys'].items()
+                    for key, value in items:
+                        if value:
+                            if value == "open_url":
+                                urwid.Button._command_map._command[key] = 'activate'
+                            value = getattr(self, "_{}".format(value))
+                        else:
+                            del self.keys[key]
+                            continue
+                        self.keys.update([(key, value)])
+                except KeyError:
+                    pass
         except FileNotFoundError:
             pass
-        # Default color palette
-        self.palettes.append([('header', 'white', 'dark blue', 'standout'),
-                              ('footer', 'white', 'dark red', 'standout'),
-                              ('search', 'white', 'dark green', 'standout'),
-                              ('msgtext', '', ''),
-                              ('msgtext:ellipses', 'light gray', 'black'),
-                              ('urlref:number:braces', 'light gray', 'black'),
-                              ('urlref:number', 'yellow', 'black', 'standout'),
-                              ('urlref:url', 'white', 'black', 'standout'),
-                              ('url:sel', 'white', 'dark blue', 'bold')])
-        # Default black & white palette
-        self.palettes.append([('header', 'black', 'light gray', 'standout'),
-                              ('footer', 'black', 'light gray', 'standout'),
-                              ('search', 'black', 'light gray', 'standout'),
-                              ('msgtext', '', ''),
-                              ('msgtext:ellipses', 'white', 'black'),
-                              ('urlref:number:braces', 'white', 'black'),
-                              ('urlref:number', 'white', 'black', 'standout'),
-                              ('urlref:url', 'white', 'black', 'standout'),
-                              ('url:sel', 'black', 'light gray', 'bold')])
-
+        try:
+            call(['xdg-open'], stdout=DEVNULL)
+            self.xdg = True
+        except OSError:
+            self.xdg = False
         self.shorten = shorten
         self.compact = compact
         self.run = run
@@ -131,6 +182,8 @@ class URLChooser:
         self.search_string = ""
         self.no_matches = False
         self.enter = False
+        self.activate_keys = [i for i, j in urwid.Button._command_map._command.items()
+                              if j == 'activate']
         self.items, self.urls = self.process_urls(extractedurls,
                                                   dedupe=dedupe,
                                                   shorten=self.shorten)
@@ -146,31 +199,34 @@ class URLChooser:
         # Store items grouped into sections
         self.items_org = grp_list(self.items)
         listbox = urwid.ListBox(self.items)
-        header = (":: q - Quit :: "
-                  "/ - search :: "
-                  "c - context :: "
-                  "C - copy to clipboard :: "
-                  "s - URL short :: "
-                  "S - all URL short :: "
-                  "g/G - top/bottom :: "
-                  "<num> - jump to <num> :: "
-                  "p - cycle palettes :: "
-                  "P - create config file ::"
-                  "u - unescape URL ::")
-        headerwid = urwid.AttrMap(urwid.Text(header), 'header')
-        self.top = urwid.Frame(listbox, headerwid)
+        self.header = (":: F1 - help/keybindings :: "
+                       "q - quit :: "
+                       "/ - search :: "
+                       "URL opening mode - {}")
+        self.link_open_modes = ["Web Browser", "Xdg-Open"] if self.xdg is True else ["Web Browser"]
+        if self.run:
+            self.link_open_modes.insert(0, self.run)
+        self.nohelp = nohelp
+        if nohelp is False:
+            self.headerwid = urwid.AttrMap(urwid.Text(
+                self.header.format(self.link_open_modes[0])), 'header')
+        else:
+            self.headerwid = None
+        self.top = urwid.Frame(listbox, self.headerwid)
         if self.urls:
             self.top.body.focus_position = \
                 (2 if self.compact is False else 0)
         self.tui = urwid.curses_display.Screen()
+        self.palette_names = list(self.palettes.keys())
         self.palette_idx = 0
         self.number = ""
+        self.help_menu = False
 
     def main(self):
         """Urwid main event loop
 
         """
-        self.loop = urwid.MainLoop(self.top, self.palettes[0], screen=self.tui,
+        self.loop = urwid.MainLoop(self.top, self.palettes[self.palette_names[0]], screen=self.tui,
                                    handle_mouse=False, input_filter=self.handle_keys,
                                    unhandled_input=self.unhandled)
         self.loop.run()
@@ -200,17 +256,20 @@ class URLChooser:
                         text = ""
                     footerwid = urwid.AttrMap(urwid.Text(text), footer)
                     self.top.footer = footerwid
-                elif k == ' ':
-                    self.search_string += " "
-                    footerwid = urwid.AttrMap(urwid.Text(text), 'footer')
-                    self.top.footer = footerwid
+                elif k in self.activate_keys:
+                    self.search_string += k
+                    self._search()
                 elif k == 'backspace':
                     self.search_string = self.search_string[:-1]
                     self._search()
-            elif k in ('enter', ' ') and self.urls and self.search is False:
-                load_text = "Loading URL..." if not self.run else "Executing: {}".format(self.run)
-                if os.environ.get('BROWSER') not in ['elinks', 'links', 'w3m', 'lynx']:
-                    self._footer_start_thread(load_text, 5)
+            elif k in self.activate_keys and \
+                    self.urls and \
+                    self.search is False and \
+                    self.help_menu is False:
+                self._open_url()
+            elif self.help_menu is True:
+                self._help_menu()
+                return []
             if k == 'up':
                 # Works around bug where the up arrow goes higher than the top list
                 # item and unintentionally triggers context and palette switches.
@@ -224,145 +283,240 @@ class URLChooser:
         return [i for i in keys if i != 'backspace']
 
     def unhandled(self, key):
-        """Add other keyboard actions (q, j, k, s, S, c, C, g, G) not handled by
-        the ListBox widget.
+        """Handle other keyboard actions not handled by the ListBox widget.
 
         """
-        size = self.tui.get_cols_rows()
+        self.key = key
+        self.size = self.tui.get_cols_rows()
         if self.search is True:
             if self.enter is False and self.no_matches is False:
                 if len(key) == 1 and key.isprintable():
                     self.search_string += key
                 self._search()
+            elif self.enter is True and not self.search_string:
+                self.search = False
+                self.enter = False
             return
-        if key in ('q', 'Q'):
-            raise urwid.ExitMainLoop()
-        elif not self.urls:
-            pass  # No other actions are useful with no URLs
-        elif key == '/':
-            if self.urls:
-                self.search = True
-                if self.compact is True:
-                    self.compact = False
-                    self.items, self.items_com = self.items_com, self.items
+        if not self.urls and key not in "Qq":
+            return  # No other actions are useful with no URLs
+        if self.help_menu is False:
+            try:
+                self.keys[key]()
+            except KeyError:
+                pass
+
+    def _quit(self):
+        """q/Q"""
+        raise urwid.ExitMainLoop()
+
+    def _open_url(self):
+        """<Enter> or <space>"""
+        load_text = "Loading URL..." if self.link_open_modes[0] != self.run \
+                else "Executing: {}".format(self.run)
+        if os.environ.get('BROWSER') not in ['elinks', 'links', 'w3m', 'lynx']:
+            self._footer_start_thread(load_text, 5)
+
+    def _help_menu(self):
+        """F1"""
+        if self.help_menu is False:
+            self.focus_pos_saved = self.top.body.focus_position
+            help_men = "\n".join(["{} - {}".format(i, j.__name__.strip('_'))
+                                  for i, j in self.keys.items() if j.__name__ !=
+                                  '_digits'])
+            help_men = "KEYBINDINGS\n" + help_men + "\n<0-9> - Jump to item"
+            docs = ("OPTIONS\n"
+                    "all_escape -- toggle unescape all URLs\n"
+                    "all_shorten -- toggle shorten all URLs\n"
+                    "bottom -- move cursor to last item\n"
+                    "clear_screen -- redraw screen\n"
+                    "clipboard -- copy highlighted URL to clipboard using xsel/xclip\n"
+                    "clipboard_pri -- copy highlighted URL to primary selection using xsel/xclip\n"
+                    "config_create -- create ~/.config/urlscan/config.json\n"
+                    "context -- show/hide context\n"
+                    "down -- cursor down\n"
+                    "help_menu -- show/hide help menu\n"
+                    "link_handler -- cycle through xdg-open, webbrowser and user-defined function\n"
+                    "open_url -- open selected URL\n"
+                    "palette -- cycle through palettes\n"
+                    "quit -- quit\n"
+                    "shorten -- toggle shorten highlighted URL\n"
+                    "top -- move to first list item\n"
+                    "up -- cursor up\n")
+            self.top.body = \
+                    urwid.ListBox(urwid.SimpleListWalker([urwid.Columns([(30, urwid.Text(help_men)),
+                                                                         urwid.Text(docs)])]))
+        else:
+            self.top.body = urwid.ListBox(self.items)
+            self.top.body.focus_position = self.focus_pos_saved
+        self.help_menu = not self.help_menu
+
+    def _search_key(self):
+        """ / """
+        if self.urls:
+            self.search = True
+            if self.compact is True:
+                self.compact = False
+                self.items, self.items_com = self.items_com, self.items
+        else:
+            return
+        self.no_matches = False
+        self.search_string = ""
+        # Reset the search highlighting
+        self._search()
+        footerwid = urwid.AttrMap(urwid.Text("Search: "), 'footer')
+        self.top.footer = footerwid
+        self.items = self.items_orig
+        self.top.body = urwid.ListBox(self.items)
+
+    def _digits(self):
+        """ 0-9 """
+        self.number += self.key
+        try:
+            if self.compact is False:
+                self.top.body.focus_position = \
+                    self.items.index(self.items_com[max(int(self.number) - 1, 0)])
             else:
-                return
-            self.no_matches = False
-            self.search_string = ""
-            # Reset the search highlighting
-            self._search()
-            footerwid = urwid.AttrMap(urwid.Text("Search: "), 'footer')
+                self.top.body.focus_position = \
+                    self.items.index(self.items[max(int(self.number) - 1, 0)])
+        except IndexError:
+            self.number = self.number[:-1]
+        self.top.keypress(self.size, "")  # Trick urwid into redisplaying the cursor
+        if self.number:
+            self._footer_start_thread("Selection: {}".format(self.number), 1)
+
+    def _clear_screen(self):
+        """ Ctrl-l """
+        self.draw_screen(self.size)
+
+    def _down(self):
+        """ j """
+        self.top.keypress(self.size, "down")
+
+    def _up(self):
+        """ k """
+        self.top.keypress(self.size, "up")
+
+    def _top(self):
+        """ g """
+        # Goto top of the list
+        self.top.body.focus_position = 2 if self.compact is False else 0
+        self.top.keypress(self.size, "")  # Trick urwid into redisplaying the cursor
+
+    def _bottom(self):
+        """ G """
+        # Goto bottom of the list
+        self.top.body.focus_position = len(self.items) - 1
+        self.top.keypress(self.size, "")  # Trick urwid into redisplaying the cursor
+
+    def _shorten(self):
+        """ s """
+        # Toggle shortened URL for selected item
+        fpo = self.top.body.focus_position
+        url_idx = len([i for i in self.items[:fpo + 1]
+                       if isinstance(i, urwid.Columns)]) - 1
+        if self.compact is False and fpo <= 1:
+            return
+        url = self.urls[url_idx]
+        short = not "..." in self.items[fpo][1].label
+        self.items[fpo][1].set_label(shorten_url(url, self.size[0], short))
+
+    def _all_shorten(self):
+        """ S """
+        # Toggle all shortened URLs
+        self.shorten = not self.shorten
+        urls = iter(self.urls)
+        for item in self.items:
+            # Each Column has (Text, Button). Update the Button label
+            if isinstance(item, urwid.Columns):
+                item[1].set_label(shorten_url(next(urls),
+                                              self.size[0],
+                                              self.shorten))
+    def _all_escape(self):
+        """ u """
+        # Toggle all escaped URLs
+        self.unesc = not self.unesc
+        self.urls, self.urls_unesc = self.urls_unesc, self.urls
+        urls = iter(self.urls)
+        for item in self.items:
+            # Each Column has (Text, Button). Update the Button label
+            if isinstance(item, urwid.Columns):
+                item[1].set_label(shorten_url(next(urls),
+                                              self.size[0],
+                                              self.shorten))
+
+    def _context(self):
+        """ c """
+        # Show/hide context
+        if self.search_string:
+            # Reset search when toggling compact mode
+            footerwid = urwid.AttrMap(urwid.Text(""), 'default')
             self.top.footer = footerwid
+            self.search_string = ""
             self.items = self.items_orig
-            self.top.body = urwid.ListBox(self.items)
-        elif key.isdigit():
-            self.number += key
-            try:
-                if self.compact is False:
-                    self.top.body.focus_position = \
-                        self.items.index(self.items_com[max(int(self.number) - 1, 0)])
-                else:
-                    self.top.body.focus_position = \
-                        self.items.index(self.items[max(int(self.number) - 1, 0)])
-            except IndexError:
-                self.number = self.number[:-1]
-            self.top.keypress(size, "")  # Trick urwid into redisplaying the cursor
-            if self.number:
-                self._footer_start_thread("Selection: {}".format(self.number), 1)
-        elif key == 'ctrl l':
-            self.draw_screen(size)
-        elif key == 'j':
-            self.top.keypress(size, "down")
-        elif key == 'k':
-            self.top.keypress(size, "up")
-        elif key == 'g':
-            # Goto top of the list
-            self.top.body.focus_position = 2 if self.compact is False else 0
-            self.top.keypress(size, "")  # Trick urwid into redisplaying the cursor
-        elif key == 'G':
-            # Goto bottom of the list
-            self.top.body.focus_position = len(self.items) - 1
-            self.top.keypress(size, "")  # Trick urwid into redisplaying the cursor
-        elif key == 's':
-            # Toggle shortened URL for selected item
-            fpo = self.top.body.focus_position
-            url_idx = len([i for i in self.items[:fpo + 1]
-                           if isinstance(i, urwid.Columns)]) - 1
-            if self.compact is False and fpo <= 1:
-                return
-            url = self.urls[url_idx]
-            short = False if "..." in self.items[fpo][1].label else True
-            self.items[fpo][1].set_label(shorten_url(url, size[0], short))
-        elif key in ('S', 'u'):
-            # Toggle all shortened or escaped URLs
-            if key == 'S':
-                self.shorten = not self.shorten
-            if key == 'u':
-                self.unesc = not self.unesc
-                self.urls, self.urls_unesc = self.urls_unesc, self.urls
-            urls = iter(self.urls)
-            for item in self.items:
-                # Each Column has (Text, Button). Update the Button label
-                if isinstance(item, urwid.Columns):
-                    item[1].set_label(shorten_url(next(urls),
-                                                  size[0],
-                                                  self.shorten))
-        elif key == 'c':
-            # Show/hide context
-            if self.search_string:
-                # Reset search when toggling compact mode
-                footerwid = urwid.AttrMap(urwid.Text(""), 'default')
-                self.top.footer = footerwid
-                self.search_string = ""
-                self.items = self.items_orig
-            fpo = self.top.body.focus_position
-            self.items, self.items_com = self.items_com, self.items
-            self.top.body = urwid.ListBox(self.items)
-            self.top.body.focus_position = self._cur_focus(fpo)
-            self.compact = not self.compact
-        elif key == 'C':
-            # Copy highlighted url to clipboard
-            fpo = self.top.body.focus_position
-            url_idx = len([i for i in self.items[:fpo + 1]
-                           if isinstance(i, urwid.Columns)]) - 1
-            if self.compact is False and fpo <= 1:
-                return
-            url = self.urls[url_idx]
+        fpo = self.top.body.focus_position
+        self.items, self.items_com = self.items_com, self.items
+        self.top.body = urwid.ListBox(self.items)
+        self.top.body.focus_position = self._cur_focus(fpo)
+        self.compact = not self.compact
+
+    def _clipboard(self, pri=False):
+        """ C """
+        # Copy highlighted url to clipboard
+        fpo = self.top.body.focus_position
+        url_idx = len([i for i in self.items[:fpo + 1]
+                       if isinstance(i, urwid.Columns)]) - 1
+        if self.compact is False and fpo <= 1:
+            return
+        url = self.urls[url_idx]
+        if pri is True:
             cmds = ("xsel -i", "xclip -i")
-            for cmd in cmds:
-                try:
-                    proc = Popen(shlex.split(cmd), stdin=PIPE)
-                    proc.communicate(input=url.encode(sys.getdefaultencoding()))
-                    self._footer_start_thread("Copied url to primary selection", 5)
-                except OSError:
-                    continue
-                break
-        elif key == 'p':
-            # Loop through available palettes
-            self.palette_idx += 1
+        else:
+            cmds = ("xsel -ib", "xclip -i -selection clipboard")
+        for cmd in cmds:
             try:
-                self.loop.screen.register_palette(self.palettes[self.palette_idx])
-            except IndexError:
-                self.loop.screen.register_palette(self.palettes[0])
-                self.palette_idx = 0
-            self.loop.screen.clear()
-        elif key == 'P':
-            # Create ~/.config/urlscan/config.json if if doesn't exist
-            if not exists(self.conf):
-                try:
-                    # Python 2/3 compatible recursive directory creation
-                    os.makedirs(dirname(expanduser(self.conf)))
-                except OSError as err:
-                    if errno.EEXIST != err.errno:
-                        raise
-                names = ["default", "bw"]
-                with open(expanduser(self.conf), 'w') as pals:
-                    pals.writelines(json.dumps(dict(zip(names,
-                                                        self.palettes)),
-                                               indent=4))
-                self._footer_start_thread("Created ~/.config/urlscan/config.json", 5)
-            else:
-                self._footer_start_thread("Config.json already exists", 5)
+                proc = Popen(shlex.split(cmd), stdin=PIPE)
+                proc.communicate(input=url.encode(sys.getdefaultencoding()))
+                self._footer_start_thread("Copied url to {} selection".format(
+                    "primary" if pri is True else "clipboard"), 5)
+            except OSError:
+                continue
+            break
+
+    def _clipboard_pri(self):
+        """ P """
+        # Copy highlighted url to primary selection
+        self._clipboard(pri=True)
+
+    def _palette(self):
+        """ p """
+        # Loop through available palettes
+        self.palette_idx += 1
+        try:
+            self.loop.screen.register_palette(self.palettes[self.palette_names[self.palette_idx]])
+        except IndexError:
+            self.loop.screen.register_palette(self.palettes[self.palette_names[0]])
+            self.palette_idx = 0
+        self.loop.screen.clear()
+
+    def _config_create(self):
+        """ --genconf """
+        # Create ~/.config/urlscan/config.json if if doesn't exist
+        if not exists(self.conf):
+            try:
+                # Python 2/3 compatible recursive directory creation
+                os.makedirs(dirname(expanduser(self.conf)))
+            except OSError as err:
+                if errno.EEXIST != err.errno:
+                    raise
+            keys = dict(zip(self.keys.keys(),
+                            [i.__name__.strip('_') for i in self.keys.values()]))
+            with open(expanduser(self.conf), 'w') as pals:
+                pals.writelines(json.dumps({"palettes": self.palettes, "keys": keys},
+                                           indent=4))
+            print("Created ~/.config/urlscan/config.json")
+        else:
+            print("~/.config/urlscan/config.json already exists")
+
 
     def _footer_start_thread(self, text, time):
         """Display given text in the footer. Clears after <time> seconds
@@ -456,6 +610,18 @@ class URLChooser:
     def _get_search(self):
         return lambda: self.search, lambda: self.enter
 
+    def _link_handler(self):
+        """Function to cycle through opening links via webbrowser module,
+        xdg-open or custom expression passed with --run.
+
+        """
+        mode = self.link_open_modes.pop()
+        self.link_open_modes.insert(0, mode)
+        if self.nohelp is False:
+            self.headerwid = urwid.AttrMap(urwid.Text(
+                self.header.format(self.link_open_modes[0])), 'header')
+            self.top.header = self.headerwid
+
     def mkbrowseto(self, url):
         """Create the urwid callback function to open the web browser or call
         another function with the URL.
@@ -481,9 +647,12 @@ class URLChooser:
                 if self._get_search()[1]() is True:
                     self.search = False
                     self.enter = False
-            elif not self.run:
+            elif self.link_open_modes[0] == "Web Browser":
                 webbrowser.open(url)
-            elif self.run and self.pipe:
+            elif self.link_open_modes[0] == "Xdg-Open":
+                run = 'xdg-open "{}"'.format(url)
+                process = Popen(shlex.split(run), stdout=PIPE, stdin=PIPE)
+            elif self.link_open_modes[0] == self.run and self.pipe:
                 process = Popen(shlex.split(self.run), stdout=PIPE, stdin=PIPE)
                 process.communicate(input=url.encode(sys.getdefaultencoding()))
             else:
